#include "utils.h"
#include "getopts.h"
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <sys/stat.h>
#ifdef _MSC_VER
#pragma warning(disable : 4800)	// int->bool conversion
#include <crtdbg.h>
#include <io.h>
#else
#include <unistd.h>
#endif

const char *GetProgramName();

int g_argc;
char **g_argv;

static time_t g_StartTime = time(0);

long GetElapsedSecs()
	{
	return (long) (time(0) - g_StartTime);
	}

FILE *g_fLog = NULL;

void Log(const char szFormat[], ...)
	{
	static bool Init = false;

	if (!Init)
		{
		const char *LogFileName = ValueOpt("log");
		if (LogFileName != 0)
			{
			g_fLog = fopen(LogFileName, "w");
			if (NULL == g_fLog)
				{
				fprintf(stderr, "Cannot open log file '%s'", LogFileName);
				exit(1);
				}
			setbuf(g_fLog, 0);
			}
		Init = true;
		}

	if (NULL == g_fLog)
		return;

	char szStr[64000];
	va_list ArgList;
	va_start(ArgList, szFormat);
	vsprintf(szStr, szFormat, ArgList);
	fprintf(g_fLog, "%s", szStr);
	fflush(g_fLog);
	}

void Die(const char szFormat[], ...)
	{
	char szStr[64000];
	va_list ArgList;
	va_start(ArgList, szFormat);
	vsprintf(szStr, szFormat, ArgList);
	Log("\n");
	Log("---Fatal error---\n");
	Log(szStr);
	Log("\n");
	Log("Elapsed time so far %ld secs\n", GetElapsedSecs());
	fprintf(stderr, "\n");
	for (int i = 0; i < g_argc; ++i)
		{
		if (i > 0)
			fprintf(stderr, " ");
		fprintf(stderr, "%s", g_argv[i]);
		}
	fprintf(stderr, "\n");
	fprintf(stderr, "*** Fatal error *** %s\n%s\n", GetProgramName(), szStr);
#ifdef _MSC_VER
	_CrtSetDbgFlag(0);
#endif
//#if _MSC_VER && _DEBUG
//	assert(false);
//#endif
	abort();
	exit(1);
	}

void *allocmem(size_t bytes)
	{
	void *p = malloc(bytes);
	if (0 == p)
		fprintf(stderr, "Out of memory allocmem(%u)", (unsigned) bytes);
	return p;
	}

void freemem(void *p)
	{
	free(p);
	}

char *strsave(const char *s)
	{
	if (0 == s)
		return 0;
	char *ptrCopy = strdup(s);
	if (0 == ptrCopy)
		fprintf(stderr, "Out of memory");
	return ptrCopy;
	}

char *mystrsave(const char *s)
	{
	if (0 == s)
		return 0;
	size_t len = strlen(s) + 1;
	char *ptrCopy = myalloc(char, len);
	memcpy(ptrCopy, s, len);
	return ptrCopy;
	}

void DeleteStdioFile(const char *FileName)
	{
	int Ok = remove(FileName);
	if (Ok != 0)
		Die("remove(%s) failed, errno=%d %s", FileName, errno, strerror(errno));
	}

FILE *OpenStdioFile(const char *FileName, bool ForWriting)
	{
	const char *Mode = (ForWriting ? "wb+" : "rb");
	FILE *f = fopen(FileName, Mode);
	if (0 == f)
		Die("Cannot open %s, errno=%d %s", FileName, errno, strerror(errno));
	return f;
	}

FILE *OpenStdioFileRW(const char *FileName)
	{
	FILE *f = fopen(FileName, "r+b");
	if (0 == f)
		Die("Cannot open %s, errno=%d %s", FileName, errno, strerror(errno));
	return f;
	}

FILE *CreateStdioFile(const char *FileName)
	{
	FILE *f = fopen(FileName, "wb+");
	if (0 == f)
		Die("Cannot create %s, errno=%d %s", FileName, errno, strerror(errno));
	return f;
	}

void SeekStdioFile(FILE *f, off_t Pos)
	{
	if (0 == f)
		Die("SeekStdioFile failed, f=NULL");
	int Ok = fseeko(f, Pos, SEEK_SET);
	off_t NewPos = ftello(f);
	if (Ok != 0 || Pos != NewPos)
		Die("SeekStdioFile(%x) failed, Ok=%d NewPos=%x",
		  (unsigned) Pos, Ok, (unsigned) NewPos);
	}

void ReadStdioFile(FILE *f, off_t Pos, void *Buffer, off_t Bytes)
	{
	if (0 == f)
		Die("ReadStdioFile failed, f=NULL");
	SeekStdioFile(f, Pos);
	off_t BytesRead = (off_t) fread(Buffer, 1, Bytes, f);
	if (BytesRead != Bytes)
		Die("ReadStdioFile failed, attempted %u bytes, read %u bytes, errno=%d",
		  (unsigned) Bytes, (unsigned) BytesRead, errno);
	}

void ReadStdioFile(FILE *f, void *Buffer, off_t Bytes)
	{
	if (0 == f)
		Die("ReadStdioFile failed, f=NULL");
	off_t BytesRead = (off_t) fread(Buffer, 1, Bytes, f);
	if (BytesRead != Bytes)
		Die("ReadStdioFile failed, attempted %u bytes, read %u bytes, errno=%d",
		  (unsigned) Bytes, (unsigned) BytesRead, errno);
	}

void WriteStdioFile(FILE *f, off_t Pos, const void *Buffer, off_t Bytes)
	{
	if (0 == f)
		Die("WriteStdioFile failed, f=NULL");
	SeekStdioFile(f, Pos);
	off_t BytesWritten = (off_t) fwrite(Buffer, 1, Bytes, f);
	if (BytesWritten != Bytes)
		Die("WriteStdioFile failed, attempted %u bytes, wrote %u bytes, errno=%d",
		  (unsigned) Bytes, (unsigned) BytesWritten, errno);
	}

void WriteStdioFile(FILE *f, const void *Buffer, off_t Bytes)
	{
	if (0 == f)
		Die("WriteStdioFile failed, f=NULL");
	off_t BytesWritten = (off_t) fwrite(Buffer, 1, Bytes, f);
	if (BytesWritten != Bytes)
		Die("WriteStdioFile failed, attempted %u bytes, wrote %u bytes, errno=%d",
		  (unsigned) Bytes, (unsigned) BytesWritten, errno);
	}

void WriteStdioFileStr(FILE *f, off_t Pos, const char *Buffer)
	{
	WriteStdioFile(f, Pos, Buffer, (off_t) strlen(Buffer));
	}

void WriteStdioFileStr(FILE *f, const char *Buffer)
	{
	WriteStdioFile(f, Buffer, (off_t) strlen(Buffer));
	}

bool ReadLineStdioFile(FILE *f, char *Line, off_t Bytes)
	{
	if (feof(f))
		return true;
	char *RetVal = fgets(Line, (int) Bytes, f);
	if (NULL == RetVal)
		{
		if (feof(f))
			return true;
		if (ferror(f))
			Die("ReadLineStdioFile: errno=%d", errno);
		Die("ReadLineStdioFile: fgets=0, feof=0, ferror=0");
		}

	if (RetVal != Line)
		Die("ReadLineStdioFile: fgets != Buffer");
	size_t n = strlen(Line);
	if (n < 1 || Line[n-1] != '\n')
		Die("ReadLineStdioFile: line too long or missing end-of-line");
	Line[n-1] = 0;
	return false;
	}

void CloseStdioFile(FILE *f)
	{
	int Ok = fclose(f);
	if (Ok != 0)
		Die("fclose(%p)=%d", f, Ok);
	}

off_t GetStdioFilePos(FILE *f)
	{
	off_t FilePos = ftello(f);
	if (FilePos < 0)
		Die("ftell=%ld", (long) FilePos);
	return FilePos;
	}

off_t GetStdioFileSize(FILE *f)
	{
	off_t CurrPos = GetStdioFilePos(f);
	int Ok = fseeko(f, 0, SEEK_END);
	if (Ok < 0)
		Die("fseek in GetFileSize");
	off_t Length = ftello(f);
	if (Length < 0)
		Die("ftell in GetFileSize");
	SeekStdioFile(f, CurrPos);
	return Length;
	}

void safestrcpy(char *s, const char *t, size_t n)
	{
	if (n == 0)
		return;

	size_t m = strlen(t);
	if (m >= n - 1)
		m = n - 1;
	memcpy(s, t, m);
	s[m] = 0;
	}

// NOT IMPLEMENTED CORRECTLY.
// Not sure if we will ever need values
// > 2GB, but at least it's localized here.
uint64 atouint64(const char *s)
	{
	return (uint64) atoi(s);
	}

coord_t StrToCoord(const char *s)
	{
	return (coord_t) atoi(s);
	}

const char *CoordToStr(coord_t Pos)
	{
	static char Str[32];
	if ((coord_t) (unsigned) Pos != Pos)
		Die("CoordToStr: Overflow");
	sprintf(Str, "%u", (unsigned) Pos);
	return Str;
	}

void *alloc(size_t bytes)
	{
	void *p = malloc(bytes);
	if (0 == p)
		Die("Out of memory, alloc(%u)\n", (unsigned) bytes);
	return p;
	}

char *strrev(char *s)
	{
	size_t L = strlen(s);
	char *save = s;
	char *t = s + L - 1;
	while (s < t)
		{
		char c = *s;
		*s++ = *t--;
		*t = c;
		}
	return save;
	}

bool streq(const char *s1, const char *s2)
	{
	return strcmp(s1, s2) == 0;
	}

void Warning(const char *Format, ...)
	{
	char Str[16000];
	va_list ArgList;
	va_start(ArgList, Format);
	vsprintf(Str, Format, ArgList);

	Log("WARNING ");
	Log(Str);
	Log("\n");

	fprintf(stderr, "\nWARNING %s\n", Str);
	}

// 4^n
uint64 pow4(unsigned n)
	{
	assert(n < 32);
	const uint64 one = 1;
	return (one << (2*n));
	}

// 20^n
uint64 pow20(unsigned n)
	{
	uint64 p = 1;
	for (unsigned i = 0; i < n; ++i)
		p *= 20;
	return p;
	}

const char *uint64tostr(uint64 i)
	{
	static char s[64];
	return uint64tostr(i, s);
	}

const char *uint64tostr(uint64 i, char *s)
	{
#ifdef 	_MSC_VER
	sprintf(s, "%I64u", i);
	return s;
#endif

#ifdef 	__GNUC__
	sprintf(s, "%llu", i);
	return s;
#endif
	Die("uint64tostr not implemented on this platform");
	return 0;
	}

const char *uint32tostr(uint32 i)
	{
	return uint64tostr(i);
	}

const char *uint32tostr(uint32 i, char *s)
	{
	return uint64tostr(i, s);
	}

bool FileExists(const char *FileName)
	{
	struct stat SD;
	int i = stat(FileName, &SD);
	return i == 0;
	}

void myassertfail(const char *Exp, const char *File, unsigned Line)
	{
	Die("%s(%u) Assertion failed: %s", File, Line, Exp);
	}

const char *atoh(const char *str)
	{
	static char s[128];

	int L = (int) strlen(str);
	if (L >= (int) sizeof(s) - 2)
		return str;

	int j = sizeof(s) - 2;
	for (int i = L-1; i >= 0; --i)
		{
		int k = L - i - 1;
		if (k%3 == 0 && k != 0)
			s[j--] = ',';
		s[j--] = str[i];
		}
	return s + j + 1;
	}

const char *UIntToStr(unsigned i)
	{
	if (i == UINT_MAX)
		return "*";

	static char s[256];
	static unsigned spos = 0;
	char *p = s + spos;
	sprintf(p, "%u", i);
	spos += 32;
	if (spos >= sizeof(s) - 32)
		spos = 0;
	return p;
	}

bool IsTerminal(int fd)
	{
	return (bool) isatty(fd);
	}

const char *SPScoreToStr(sp_score_t Score)
	{
	static char Str[32];
	sprintf(Str, "%d", (int) Score);
	return Str;
	}

const char *ScoreToStr(score_t Score)
	{
	static char Str[32];
#if INTEGER_DP
	sprintf(Str, "%d", (int) Score);
#else
	sprintf(Str, "%g", (double) Score);
#endif
	return Str;
	}
